#!/usr/bin/python3

"""
Extract data from a csv file for use in a sankey visualization

Produce a json file

Assume all nodes have distinct names, also between different levels.
"""

infile = 'sk-joh.csv'
outfile = 'data-joh.json'
column_l1 = 2
column_l2 = 1
column_l3 = 0
column_n = 5

threshold = 0.005

max_total_nodes = 90

add_other = True

import json
from collections import Counter

threshold_l1 = threshold
threshold_l2 = threshold
threshold_l3 = threshold

nodes = []
links = []

with open(infile, 'r') as infile:
    lines = infile.read().strip().split('\n')

n12 = Counter()
n23 = Counter()
n1 = Counter()
n2 = Counter()
n3 = Counter()

nodes1 = []
nodes2 = []
nodes3 = []

nodes1_children = {}
nodes2_children = {}

n_tot = 0
for line in lines:
    f = line.split('\t')
    n = int(f[column_n])
    if n > 0:
        l1 = f[column_l1]
        nodes1.append(l1)
        l2 = f[column_l2]
        if l2 == 'General Science & Technology':
            l2 = 'General Science & Techn.'
        nodes2.append(l2)
        l3 = f[column_l3]
        nodes3.append(l3)
        n23[(l1, l2, l3)] += n
        n12[(l1, l2)] += n
        n1[l1] += n
        n2[l2] += n
        n3[l3] += n
        n_tot += n
        if l1 not in nodes1_children:
            nodes1_children[l1] = set([])
        nodes1_children[l1] |= set([l2])
        if l2 not in nodes2_children:
            nodes2_children[l2] = set([])
        nodes2_children[l2] |= set([l3])

#print(n1)
#print()
#print(n2)
#print()
#print(n3)

nodes1 = list(set(nodes1))
nodes2 = list(set(nodes2))
nodes3 = list(set(nodes3))
nodes1.sort()
nodes2.sort()
nodes3.sort()

# select nodes
nodes1_sel = []
for node1 in nodes1:
    if n1[node1] >= threshold_l1 * n_tot:
        nodes1_sel.append(node1)
nodes2_sel = []
for node2 in nodes2:
    if n2[node2] >= threshold_l2 * n_tot:
        nodes2_sel.append(node2)
max_count_l3 = max_total_nodes - len(nodes1) - len(nodes2)
nodes3_cand = list(map(lambda u: u[0], n3.most_common(max_count_l3)))
nodes3_sel = []
for node3 in nodes3_cand:
    if n3[node3] >= threshold_l3 * n_tot:
        nodes3_sel.append(node3)

#print(nodes1_sel)
#print()
#print(nodes2_sel)
#print()
#print(nodes3_sel)

# restrict nodes2_sel to those which have at least one child
nodes2_sel = [l2 for l2 in nodes2_sel if set(nodes2_children[l2]) & set(nodes3_sel)]
# restrict nodes1_sel to those which have at least one child
nodes1_sel = [l1 for l1 in nodes1_sel if set(nodes1_children[l1]) & set(nodes2_sel)]
nodes = nodes1_sel + nodes2_sel + nodes3_sel

if add_other:
    if len(nodes1_sel) < len(nodes1):
        v = sum([n1[l1] for l1 in nodes1 if l1 not in nodes1_sel])
        nodes.append('Other')
        nodes.append('Other ')
        nodes.append('Other  ')
        o1 = nodes.index('Other')
        o2 = nodes.index('Other ')
        o3 = nodes.index('Other  ')
        links.append({'source': o1, 'target': o2, 'value': v})
        links.append({'source': o2, 'target': o3, 'value': v})
    for l1 in nodes1_sel:
        n2_sum = sum([n2[l2] for l2 in nodes1_children[l1] & set(nodes2_sel)])
        if n1[l1] > n2_sum:
            v = n1[l1] - n2_sum
            nodes.append(l1 + ' / Other')
            s = nodes.index(l1)
            t = nodes.index(l1 + ' / Other')
            links.append({'source': s, 'target': t, 'value': v})
            nodes.append(l1 + ' / Other ')
            s = nodes.index(l1 + ' / Other')
            t = nodes.index(l1 + ' / Other ')
            links.append({'source': s, 'target': t, 'value': v})
    for l2 in nodes2_sel:
        n3_sum = sum([n3[l3] for l3 in nodes2_children[l2] & set(nodes3_sel)])
        if n2[l2] > n3_sum:
            v = n2[l2] - n3_sum
            nodes.append(l2 + ' / Other')
            s = nodes.index(l2)
            t = nodes.index(l2 + ' / Other')
            links.append({'source': s, 'target': t, 'value': v})

print('Using %s nodes.' % len(nodes))

for l1, l2 in n12:
    if l1 in nodes1_sel and l2 in nodes2_sel:
        s = nodes.index(l1)
        t = nodes.index(l2)
        v = n12[(l1, l2)]
        links.append({'source': s, 'target': t, 'value': v}) 

for l1, l2, l3 in n23:
    if l1 in nodes1_sel and l2 in nodes2_sel and l3 in nodes3_sel:
        s = nodes.index(l2)
        t = nodes.index(l3)
        v = n23[(l1, l2, l3)]
        links.append({'source': s, 'target': t, 'value': v})

res = {'nodes': [{'name': n} for n in nodes], 'links': links}

with open(outfile, 'w') as outfile:
    outfile.write(json.dumps(res, separators=(',', ':'), sort_keys=False).replace('},', '},\n').replace('}],', '}],\n'))

